use abi_stable::{
    declare_root_module_statics,
    external_types::crossbeam_channel::RSender,
    library::RootModule,
    package_version_strings, sabi_trait,
    sabi_types::VersionStrings,
    std_types::{RBox, RResult, RStr, RVec},
    StableAbi,
};

use crate::chacha::ffi_value::FfiValue;

pub mod error;
pub use error::Error;

#[sabi_trait]
/// A plugin which is loaded by the application,and provides some functionality.
// #[sabi(debug_print)]
pub trait Plugin: Clone + Debug + Display + Send + Sync {
    /// Invoke a function in the plugin.
    ///
    /// # Parameters
    /// - `module`: The module in which the function is defined.
    /// - `ty`: The type of the function.
    /// - `name`: The name of the function.
    /// - `args`: The arguments to the function.
    ///
    /// Invoke a function in the plugin, using the `module`, `ty`, and `name` to
    /// locate the function, and `args` as the arguments to the function.
    fn invoke_func(
        &self,
        module: RStr<'_>,
        ty: RStr<'_>,
        name: RStr<'_>,
        args: RVec<FfiValue>,
    ) -> RResult<FfiValue, Error>;

    /// Invoke a mutable function in the plugin.
    ///
    /// # Parameters
    /// - `module`: The module in which the function is defined.
    /// - `ty`: The type of the function.
    /// - `name`: The name of the function.
    /// - `args`: The arguments to the function.
    ///
    /// Invoke a mutable function in the plugin, using the `module`, `ty`, and `name` to
    /// locate the function, and `args` as the arguments to the function.
    ///
    /// This is the same as `invoke_func` but with a mutable reference to self.
    fn invoke_func_mut(
        &mut self,
        module: RStr<'_>,
        ty: RStr<'_>,
        name: RStr<'_>,
        args: RVec<FfiValue>,
    ) -> RResult<FfiValue, Error>;

    /// The `#[sabi(last_prefix_field)]` attribute here means that this is the last method
    /// that was defined in the first compatible version of the library
    /// (0.1.0, 0.2.0, 0.3.0, 1.0.0, 2.0.0 ,etc),
    /// requiring new methods to always be added below preexisting ones.
    ///
    /// The `#[sabi(last_prefix_field)]` attribute would stay on this method until the library
    /// bumps its "major" version,
    /// at which point it would be moved to the last method at the time.
    #[sabi(last_prefix_field)]
    /// Return the name
    ///
    /// The name of this plugin.
    fn name(&self) -> RStr<'_>;
}

pub type PluginType = Plugin_TO<'static, RBox<()>>;

/// A Lambda invocation
///
/// This little beauty encompasses everything you need to invoke a dwarf lambda
/// from a dynamic library.
///
#[repr(C)]
#[derive(StableAbi)]
pub struct LambdaCall {
    /// A pointer into the table of lambda functions.
    pub lambda: usize,
    /// Arguments to the lambda function.
    pub args: RVec<FfiValue>,
    /// A channel to send the result of the lambda function.
    pub result: RSender<RResult<FfiValue, Error>>,
}

/// The root module of a`plugin` dynamic library.
///
#[repr(C)]
#[derive(StableAbi)]
#[sabi(kind(Prefix(prefix_ref = PluginModRef)))]
#[sabi(missing_field(panic))]
pub struct PluginModule {
    pub name: extern "C" fn() -> RStr<'static>,
    /// Constructs the plugin.
    ///
    ///
    /// The `#[sabi(last_prefix_field)]` attribute here means that this is the last field in this struct
    /// that was defined in the first compatible version of the library
    /// (0.1.0, 0.2.0, 0.3.0, 1.0.0, 2.0.0 ,etc),
    /// requiring new fields to always be added below preexisting ones.
    ///
    /// The `#[sabi(last_prefix_field)]` attribute would stay on this field until the library
    /// bumps its "major" version,
    /// at which point it would be moved to the last field at the time.
    ///
    #[sabi(last_prefix_field)]
    pub new: extern "C" fn(RSender<LambdaCall>, RVec<FfiValue>) -> RResult<PluginType, Error>,
}

impl RootModule for PluginModRef {
    declare_root_module_statics! {PluginModRef}
    const BASE_NAME: &'static str = "plugin";
    const NAME: &'static str = "plugin";
    const VERSION_STRINGS: VersionStrings = package_version_strings!();
}
